#!/usr/bin/env python
# encoding: utf-8
"""
zwcontroller.py

Created by Graeme Sutherland on 2008-03-10.
Copyright (c) 2008 TrickleStar Limited. All rights reserved.
"""

import sys
import os
import unittest
import zwconstants
from zwbasecommand import *
from zwserialapi import *
from zwtransport import fid

class ZWGetControllerCapabilities(ZWBaseCommand):
    """
    Capabilities of the controller we are talking to at the end
    of the serial interface
    HOST->ZW: REQ | 0x05 
    ZW->HOST: RES | 0x05 | RetVal
    """
    def __init__(self, serial_api):
        ZWBaseCommand.__init__(self, serial_api)
        self.name='ZW_GetControllerCapabilities'
        self.capabilities = None
        self.wait_for_response = True
        self.match=[[RES, 0x05]]
        self.frame = ZWFrame(REQ)
        self.frame.set_command(0x05)
    
    def get_capabilities(self):
        self.send()
        return self._decode_capabilities()
                
    def _decode_capabilities(self):
        """
        decode return value: 
            we need and ACK and a resonable response of the correct type and a return value
        """
        if not self.frame.ack:
            raise ZWCommandFailedError, "No reply to capabilities (no ACK)"
        logging.debug(' cap frame is ' + str(self.frame.responses[0]))
        self.capabilities = self.frame.get_response_byte(1)
        logging.debug ('capabilities is %d', self.capabilities)    
        return ControllerCapabilities(self.capabilities)

class ControllerCapabilities(object):
    
    texts = {
                CONTROLLER_IS_SECONDARY: 'secondary',
                CONTROLLER_ON_OTHER_NETWORK: 'on other network',
                CONTROLLER_IS_SUC: 'suc',
                CONTROLLER_NODEID_SERVER_PRESENT: 'inclusion',
                CONTROLLER_IS_REAL_PRIMARY: 'primary'
                }
    
    
    def __init__(self, cap):
        self.capability = cap
    
    def is_secondary(self):
        return (self.capability & CONTROLLER_IS_SECONDARY) != 0

    def is_on_other_network(self):
        return (self.capability & CONTROLLER_ON_OTHER_NETWORK) != 0

    def is_suc(self):
        return (self.capability & CONTROLLER_IS_SUC) != 0
        
    def is_inclusion(self):
        return (self.capability & CONTROLLER_NODEID_SERVER_PRESENT) != 0
    
    def is_real_primary(self):
        return (self.capability & CONTROLLER_IS_REAL_PRIMARY) != 0

    def __str__(self):
        caps = []
        for f in self.texts.keys():
            if (f & self.capability) != 0:
                caps.append(self.texts[f])
        return ",".join(caps) 


class ZWControllerTests(unittest.TestCase):
    def setUp(self):
        pass


class ZWSetLearnMode(ZWBaseCommand):
    """
    Set the USB device into learn mode so it can be made a controller
    HOST->ZW: REQ | 0x50 | mode | funcID 
    ZW->HOST: REQ | 0x50 | funcID | bStatus | bSource | bLen | pCmd[] 
    """
    def __init__(self, serial_api):
        ZWBaseCommand.__init__(self, serial_api)
        self.name='ZW_SetLearnMode'
        self.wait_for_response = False
        self.frame = ZWFrame(REQ)
        self.match = []
    
    def start_learn(self):
        self.frame.set_command([0x50, 0x01, 0xe1])
        self.send()
        if not self.frame.ack:
            raise ZWCommandFailedError, "No ACK for command"
    
    def stop_learn(self):
        self.frame.set_command([0x50, 0x0, 0xe0])
        self.send()
        if not self.frame.ack:
            raise ZWCommandFailedError, "No ACK for command"
    
    
    
class ZWVersion(ZWBaseCommand):
    """
    ZW_VERSION
    HOST->ZW: REQ | 0x15 
    ZW->HOST: RES | 0x15 | buffer (12 bytes) | library type 
    buffer has:
    ZW->HOST: RES | 0x07 | SERIAL_APPL_VERSION | SERIAL_APPL_REVISION | 
    SERIALAPI_MANUFACTURER_ID1 | SERIALAPI_MANUFACTURER_ID2 | 
    SERIALAPI_MANUFACTURER_PRODUCT_TYPE1 | 
    SERIALAPI_MANUFACTURER_PRODUCT_TYPE2 | 
    SERIALAPI_MANUFACTURER_PRODUCT_ID1 | SERIALAPI_MANUFACTURER_PRODUCT_ID2 | 
    FUNCID_SUPPORTED_BITMASK[] 
    SERIAL_APPL_VERSION is the Serial API application Version number. 
    SERIAL_APPL_REVISION is the Serial API application Revision number. 
    SERIALAPI_MANUFACTURER_ID1 is the Serial API application manufacturer_id (MSB). 
    SERIALAPI_MANUFACTURER_ID2 is the Serial API application manufacturer_id (LSB). 
    SERIALAPI_MANUFACTURER_PRODUCT_TYPE1 is the Serial API application manufacturer product 
    type (MSB). 
    SERIALAPI_MANUFACTURER_PRODUCT_TYPE2 is the Serial API application manufacturer product 
    type (LSB). 
    SERIALAPI_MANUFACTURER_PRODUCT_ID1 is the Serial API application manufacturer product id 
    (MSB). 
    SERIALAPI_MANUFACTURER_PRODUCT_ID2 is the Serial API application manufacturer product id 
    (LSB). 
    FUNCID_SUPPORTED_BITMASK[] is a bitmask where every Serial API function ID which is supported 
    has a corresponding bit in the bitmask set to ‘1’. All Serial API function IDs which are not supported 
    have their corresponding bit set to ‘0’. First byte in bitmask corresponds to FuncIDs 1-8 where bit 0 
    corresponds to FuncID 1 and bit 7 corresponds to FuncID 8. Second byte in bitmask then corresponds 
    to FuncIDs 9-16 and so on. 
    
    
    
    
    """
    def __init__(self, serial):
        ZWBaseCommand.__init__(self, serial)
        self.name = 'ZW_Version'
        self.wait_for_response = True
        self.frame = ZWFrame(REQ)
        self.frame.set_command(0x15)
        self.match = [[RES, Ox02]]
        
    def get_version(self):
        self.send()
        self._decode_version()
        return self.version_string
    
    def get_version_detail(self):
        self.send()
        return VersionDetail(self.frame)
    
    def _decode_version(self):
        if not self.frame.ack:
            raise ZWCommandFailedError, "No ACK for command"
        vstring = self.frame.get_response_string(1,12)
        lib_type = self.frame.get_response_byte(13)
        self.version_string= vstring + ' Library type %s' % hex(lib_type)

class VersionDetail(object):
    def __init__(self, frame):
        if not self.frame.ack:
            raise ZWCommandFailedError, "No ACK for command"
        #TODO... not currently used


class ZWSpecialVersion(ZWBaseCommand):
    """
    HOST->ZW: REQ | 0x02 
    (Controller) ZW->HOST: RES | 0x02 | ver | capabilities | 29 | nodes[29] | chip_type | chip_version 
    (Slave) ZW->HOST: RES | 0x02 | ver | capabilities | 0 | chip_type | chip_version 
    Nodes[29] is a node bitmask. 
    Capabilities flag: 
    Bit 0: 0 = Controller API; 1 = Slave API 
    Bit 1: 0 = Timer functions not supported; 1 = Timer functions supported. 
    Bit 2: 0 = Primary Controller; 1 = Secondary Controller 
    Bit 3-7: reserved
    """
    def __init__(self, serial):
        ZWBaseCommand.__init__(self, serial)
        self.name = 'ZW_Version (special 0x02)'
        self.wait_for_response = True
        self.frame = ZWFrame(REQ)
        self.frame.set_command(0x02)
        self.match=[[RES, 0x02]]
    
    def get_special_version(self):
        self.send()
        self._decode_special_version()
        return self.version_string
    
    def _decode_special_version(self):
        if not self.frame.ack:
            raise ZWCommandFailedError, "No ACK for command"
        ver = self.frame.get_response_byte(1)
        cap = self.frame.get_response_byte(2)
        node_mask_count = self.frame.get_response_byte(3)
        node_masks = [self.frame.get_response_byte(4+n) for n in range(node_mask_count)]
        chip_type = self.frame.get_response_byte(4 + node_mask_count)
        chip_version = self.frame.get_response_byte(5 + node_mask_count)
        self.version_string= 'Special Version info: ver=' + hex(ver) + ' cap=' + hex(cap) + ' node_count=' + str(node_mask_count) + ' node_mask=[' + ','.join([hex(n) for n in node_masks]) +  '] chip_type=' + hex(chip_type) + ' chip_version=' + hex(chip_version)
        
    def get_node_list(self):
        self.send()
        return self.nodes_list()
    
    def nodes_list(self):
        if not self.frame.ack:
            raise ZWCommandFailedError, "No ACK for command"
        for r in self.frame.responses:
            if r.command[0] == 0x02:
                node_list=[]
                node_mask_count = r.command[3]
                node_masks = [r.command[4+n] for n in range(node_mask_count)]
                mask_index = 0
                for n in node_masks:
                    for b in range(8):
                        #logging.debug ("n = %d b = %d mask_index=%d" % (n, b, mask_index))
                        if n & ( 1 << b):
                            logging.debug('got node at ' + str((mask_index*8)+b+1))
                            node_list.append((mask_index * 8) + b + 1)
                    mask_index += 1
                return node_list
        raise ZWCommandFailedError, "No response"


class ZWGetNodeProtocolInfo(ZWBaseCommand):
    """
    Retrieve information from the eeprom of the controller.  So, this is guaranteed to give us
    a unit for something in the node list
    
    HOST->ZW: REQ | 0x41 | bNodeID 
    ZW->HOST: RES | 0x41 | nodeInfo (see figure 6) 
    response:[0x41,0xc9,0x4,0x0,0x3,0x11,0x0]
    
    """

    def __init__(self, serial, node_id):
        """ 
            give serial port and node id
        """    
        ZWBaseCommand.__init__(self, serial)
        self.name = 'ZWGetNodeProtocolInfo'
        self.node_id = node_id
        self.wait_for_response = True
        self.frame = ZWFrame(REQ)
        self.frame.set_command([0x41, node_id])
        self.match=[[RES,0x41]]
        
    def get_protocol_info(self):
        self.send()
        if not self.frame.ack:
            raise ZWCommandFailedError, "No ACK for command"
        self._decode_protocol_info()
        return self.protocol_string
        
    def _decode_protocol_info(self):
        capability = self.frame.get_response_byte(1)
        listening = (capability & 0x80) > 0
        capability &= 0x7f
        security = self.frame.get_response_byte(2)
        opt_func = (security & 0x80) > 0
        security &= 0x7f
        basic = self.frame.get_response_byte(4)
        generic = self.frame.get_response_byte(5)
        specific = self.frame.get_response_byte(6)
        
        self.protocol_string = "node=%d capability=%s listening=%s security=%s opt=%s basic=%s generic=%s specific=%s" % (self.node_id, hex(capability), listening, hex(security), opt_func, hex(basic), hex(generic), hex(specific))

    def get_protocols(self):
        self.send()
        if not self.frame.ack:
            raise ZWCommandFailedError, "No ACK for command"
        for r in self.frame.responses:
            if r.command[0] == 0x41:
                resp = r.command
                capability = resp[1]
                listening = (capability & 0x80) > 0
                capability &= 0x7f
                security = resp[2]
                opt_func = (security & 0x80) > 0
                security &= 0x7f
                basic = resp[4]
                generic = resp[5]
                specific = resp[6]
                return Protocols(basic, generic, specific, [])
        raise ZWCommandFailedError, "No response"


class ZWRequestNodeInfo(ZWBaseCommand):
    """
    Go to the device and ask for node info, including specific command classes... 
    
    response
    ZW->HOST: REQ | 0x49 | bStatus | bNodeID | bLen | basic | generic | specific | commandclasses[] 
    
    """
    def __init__(self, serial, node_id):
        """docstring for __init__"""    
        ZWBaseCommand.__init__(self, serial)
        self.name = 'ZWRequestNodeInfo'
        self.node_id = node_id
        self.wait_for_response = True
        self.frame = ZWFrame(REQ)
        self.frame.set_command([0x60, node_id]) 
        self.match=[[RES,0x60],[REQ,0x49]]

    def get_protocol_info(self):
        self.send()
        if not self.frame.ack:
            raise ZWCommandFailedError, "No ACK for command"
        self.frame.responses = []
        self.serial.read_something(self.frame)
        self._decode_protocol_info()
        return self.protocol_string
           
    def get_protocols(self):
        self.send()
        if not self.frame.ack:
            raise ZWCommandFailedError, "No ACK for command"
        # get the correct response here
        self.frame.responses = []
        self.serial.read_something(self.frame)
        if len(self.match[1]) < 3:
            return None #nothing found
        f = self.match[1][2]
        if f.command[0] == 0x49:
            if f.command[1] != 0x84: return None
            mylen = f.command[3]
            basic = f.command[4]
            generic = f.command[5]
            specific = f.command[6]
            command_classes = [ f.command[7+i] for i in range(mylen-4) ]
            return Protocols(basic, generic, specific, command_classes)
        else:
            return None

    def _decode_protocol_info(self):
        status = self.frame.get_response_byte(1)
        nodeid = self.frame.get_response_byte(2)
        mylen = self.frame.get_response_byte(3)
        basic = self.frame.get_response_byte(4)
        generic = self.frame.get_response_byte(5)
        specific = self.frame.get_response_byte(6)
        command_classes = [ self.frame.get_response_byte(7+i) for i in range(mylen-4) ]
        self.protocol_string = "node=%d status=%s basic=%s generic=%s specific=%s" % (self.node_id, hex(status), hex(basic), hex(generic), hex(specific))
        self.protocol_string += " Command classes: [" + ",".join([ hex(c) for c in command_classes]) + ']'

class Protocols(object):
    def __init__(self, basic, generic, specific, commands):
        self.basic = basic
        self.generic = generic
        self.specific = specific
        self.command_classes = commands
    
    def __str__(self):
        protocol_string = "basic=%s generic=%s specific=%s" % (hex(self.basic), hex(self.generic), hex(self.specific))
        protocol_string += " Command classes: [" + ",".join([ hex(c) for c in self.command_classes]) + ']'
        return protocol_string

class ZWRemoveNodeFromNetwork(ZWBaseCommand):
    """ZWRemoveNodeFromNetwork
    HOST->ZW: REQ | 0x4B | mode | funcID 
    ZW->HOST: REQ | 0x4B | funcID | bStatus | bSource | bLen | basic | generic | specific | cmdclasses[] 
    """
    REMOVE_NODE_ANY = 0x01
    REMOVE_NODE_STOP = 0x05
    
    def __init__(self, serial):
        ZWBaseCommand.__init__(self, serial)
        self.name='ZWRemoveNodeFromNetwork'
        self.wait_for_response = False
        self.match=[]
    
    def start_remove(self):
        self.frame = ZWFrame(REQ)
        self.frame.set_command([0x4B, self.REMOVE_NODE_ANY, 0x9a])
        self.send()
        if not self.frame.ack:
            raise ZWCommandFailedError, "No ACK for command"

    def stop_remove(self):
        self.frame = ZWFrame(REQ)
        self.frame.set_command([0x4B, self.REMOVE_NODE_STOP, 0x9a])
        self.send()
        if not self.frame.ack:
            raise ZWCommandFailedError, "No ACK for command"
        

class ZWAddNodeToNetwork(ZWBaseCommand):
    """ZWAddNodeToNetwork

    """
    ADD_NODE_ANY        =  1
    ADD_NODE_CONTROLLER =  2
    ADD_NODE_SLAVE      =  3
    ADD_NODE_EXISTING   =  4
    ADD_NODE_STOP       =  5
    ADD_NODE_STOP_FAILED = 6

    def __init__(self, serial):
        ZWBaseCommand.__init__(self, serial)
        self.name='ZWAddNodeFromNetwork'
        self.wait_for_response = False
        self.match=[]

    def start_add(self, replicate=False):
        if replicate:
            opts = self.ADD_NODE_CONTROLLER
        else:
            opts = self.ADD_NODE_ANY
        self.frame = ZWFrame(REQ)
        self.frame.set_command([0x4A, self.ADD_NODE_ANY, 0xaa])
        self.send()
        if not self.frame.ack:
            raise ZWCommandFailedError, "No ACK for command"

    def stop_add(self):
        self.frame = ZWFrame(REQ)
        self.frame.set_command([0x4A, self.ADD_NODE_STOP, 0xaa])
        self.send()
        if not self.frame.ack:
            raise ZWCommandFailedError, "No ACK for command"


class MemoryGetID(ZWBaseCommand):
    """docstring for MemoryGetID """
    def __init__(self, serial):
        ZWBaseCommand.__init__(self,serial)
        self.name = 'MemoryGetID'
        self.wait_for_response = True
        self.frame = ZWFrame(REQ)
        self.frame.set_command(0x20)
        self.match=[[RES, 0x20]]
    
    def get_ids(self):
        self.send()
        if not self.frame.ack:
            raise ZWCommandFailedError, "No ACK for command"
        return self._decode_ids()
    
    def _decode_ids(self):

        for r in self.frame.responses:
            if (r.command[0] == 0x20): 
                home_id = 0;
                for i in range(4):
                    home_id = home_id * 0x100;
                    home_id += r.command[i+1]
                node_id = r.command[5]
                return (home_id, node_id)
        raise ZWCommandFailedError, "No response"


class ZWSetDefault(ZWBaseCommand):
    """
    Set controller (USB interface) to defaults. This is network knackering
    HOST->ZW: REQ | 0x42 | funcID 
    ZW->HOST: REQ | 0x42 | funcID 
    
    """
    def __init__(self, serial):
        ZWBaseCommand.__init__(self,serial)
        self.name = 'ZWSetDefault'
        self.wait_for_response = True
        self.frame = ZWFrame(REQ)
        self.frame.set_command(0x42)
        self.match=[[REQ, 0x42]]

    def go(self):
        self.send()
        if not self.frame.ack:
            raise ZWCommandFailedError, "No ACK for command"


class ApplicationNodeInformation(ZWBaseCommand):
    """
    Send node info to the USB device
    serial API only
    HOST->ZW: REQ | 0x03 | deviceOptionsMask | generic | specific | parmLength | nodeParm[] 
    device options:
    #define APPLICATION_NODEINFO_NOT_LISTENING            0x00
    #define APPLICATION_NODEINFO_LISTENING                0x01
    #define APPLICATION_NODEINFO_OPTIONAL_FUNCTIONALITY   0x02
    
    """
    def __init__(self, serial):
        ZWBaseCommand.__init__(self,serial)
        self.name = 'ApplicationNodeInformation'
        self.wait_for_response = False
        self.frame = ZWFrame(REQ)
        self.match=[]
        #self.frame.set_command(0x03)
        
    
    def go(self, is_listening, has_optional_func, generic, specific, classes):
        # store classes into message
        if is_listening:
            opt = 0x01
        else:
            opt = 0x00
        if has_optional_func:
            opt |= 0x02
        self.frame.set_command([0x03, opt, generic, specific, len(classes)] + classes)
        self.send()
        if not self.frame.ack:
            raise ZWCommandFailedError, "No ACK for command"


class ZWRequestNetworkUpdate(ZWBaseCommand):
    """
    
    Request updates from the network when not the primary controller or, when
    a secondary controller:
    HOST->ZW: REQ | 0x53 | funcID 
    ZW->HOST: RES | 0x53 | retVal 
    ZW->HOST: REQ | 0x53 | funcID | txStatus 
    
    
    """
    def __init__(self, serial):
        ZWBaseCommand.__init__(self,serial)
        self.name = 'RequestNetworkUpdate'
        self.wait_for_response = True
        self.frame = ZWFrame(REQ)
        self.frame.set_command([0x53, 0xee])
        self.match=[[RES,0x53]]
    
    def go(self):
        self.send()
        if not self.frame.ack:
            raise ZWCommandFailedError, "No ACK for command"
        for r in self.frame.responses:
            if (r.command[0] == 0x53): 
                status = r.command[1]
                if status != 0:
                    return [True]
                else:
                    return [False]
        return [False]

class ZWGetSUCNodeID(ZWBaseCommand):
    """
    Used to get the currently registered SUC node ID. 
    Defined in: ZW_controller_api.h 
    Return value: 
    BYTE The node ID (1..232) on the currently 
    registered SUC, if ZERO then no SUC 
    available. 
    
    Called if secondary or inclusion so we can find out who is in charge

    Serial API: 
    HOST->ZW: REQ | 0x56 
    ZW->HOST: RES | 0x56 | SUCNodeID
    
    """
    def __init__(self, serial):
        ZWBaseCommand.__init__(self,serial)
        self.name = 'SetSUCNodeID'
        self.wait_for_response = True
        self.frame = ZWFrame(REQ)
        self.match=[[RES,0x56]]
    
    def get_id(self):
        self.frame.set_command([0x56])
        self.send()
        if not self.frame.ack:
            raise ZWCommandFailedError, "No ACK for command"
        # extract result
        for r in self.frame.responses:
            if len(r.command) >=2 and r.command[0]==0x56:
                return r.command[1]
        raise ZWCommandFailedError, "No response to ZWGetSUCNodeID"

class ZWSetSUCNodeID(ZWBaseCommand):
    """
    Used to configure a static/bridge controller to be a SUC/SIS node or not. The primary controller should 
    use this function to set a static/bridge controller to be the SUC/SIS node, or it could be used to stop 
    previously chosen static/bridge controller being a SUC/SIS node. 
    HOST->ZW: REQ | 0x54 | nodeID | SUCState | bTxOption | capabilities | funcID 
    ZW->HOST: RES | 0x54 | RetVal 
    ZW->HOST: REQ | 0x54 | funcID | txStatus 
    In case ZW_SetSUCNodeID is called locally with the controllers own node ID then only the response is 
    returned. In case true is returned in the response then it can be interpreted as the command is now 
    executed successfully. 
    

    
    """
    ZW_SUC_FUNC_BASIC_SUC      = 0x00
    ZW_SUC_FUNC_NODEID_SERVER  = 0x01
    
    def __init__(self, serial):
        ZWBaseCommand.__init__(self,serial)
        self.name = 'SetSUCNodeID'
        self.wait_for_response = True
        self.frame = ZWFrame(REQ)
        self.match = [[RES, 0x54],[REQ,0x54]]
    
    def go(self, node_id, suc_state=1, tx_option=0, capabilities=ZW_SUC_FUNC_BASIC_SUC):
        fi = fid.get_fid(node_id)
        self.frame.set_command([0x54, node_id, suc_state, tx_option, capabilities, fi])
        self.send()
        if not self.frame.ack:
            raise ZWCommandFailedError, "No ACK for command"
        
    
    
class ZWEnableSUC(ZWBaseCommand):
    """
    Enable/disable SUC functions in this controller
    HOST->ZW: REQ | 0x52 | state | capabilities 
    ZW->HOST: RES | 0x52 | retVal 
    """
    ZW_SUC_FUNC_BASIC_SUC       = 0x00
    ZW_SUC_FUNC_NODEID_SERVER   = 0x01
    
    def __init__(self, serial):
        ZWBaseCommand.__init__(self,serial)
        self.name = 'EnableSUC'
        self.wait_for_response = True
        self.frame = ZWFrame(REQ)
        self.match =[[RES,0x52]]
    
    def enable_suc(self, enable=True, capabilities=ZW_SUC_FUNC_BASIC_SUC):
        if enable:
            en = 1
        else:
            en = 0
        self.frame.set_command([0x52, en, capabilities])
        self.send()
        if not self.frame.ack:
            raise ZWCommandFailedError, "No ACK for command"
    
    
class ZWReplicationReceiveComplete(ZWBaseCommand):
    """
    Say replication ok
    HOST->ZW: REQ | 0x44
    """

    def __init__(self, serial):
        ZWBaseCommand.__init__(self,serial)
        self.name = 'ReplicationReceiveComplete'
        self.wait_for_response = False
        self.frame = ZWFrame(REQ)
        self.match=[]
        
    def go(self):
        self.frame.set_command(0x44)
        self.send()

class SerialAPISoftReset(ZWBaseCommand):
    def __init__(self, serial):
        ZWBaseCommand.__init__(self,serial)
        self.name = 'SerialAPISoftReset'
        self.wait_for_response = False
        self.frame = ZWFrame(REQ)
        self.match=[]
        
    def go(self):
        self.frame.set_command(0x08)
        self.send()



if __name__ == '__main__':
    unittest.main()