#!/usr/bin/env python
# encoding: utf-8
"""
zwmodel.py

classes for modelling ZWave (rather than just commands, these help to model
and operate networks)


Created by Graeme Sutherland on 2008-04-24.
Copyright (c) 2008 TrickleStar Limited. All rights reserved.
"""
import pdb
import sys
import os
import unittest
import exceptions
import logging
import threading
from Queue import Queue
from Queue import Empty
import time
import sched
import random

from zwserialapi import *
from zwbasecommand import *
from zwcontroller import *
from zwmodel import *
from zwclassresolver import *
from zwconfig import *
from devicelog import DeviceLog
from manufacturerspecific import ManufacturerSpecific
from notify import Notifier
from clientcommand import ClientCommand
from idrange import check_id

class ZWModel(object):
    """
    
    A model has a network and some devices of different kinds
    
    """
    def __init__(self, log):
        self.network = None
        self.devices = {}
        self.state = 'unknown'
        self.send_queue = Queue(50)
        self.last_minute=-1
        self.last_poll_time = 0
        self.last_network_update_time = 0
        self.network_update_delay = 24*60*60
        self.short_network_update_delay = 15*60
        self.network_update_count = 0
        self.poll_lock = threading.Lock()
        self.startup = False
        self.mode="run"
        self.mode_end = 0    # seconds
        self.mode_delay = 15 # seconds
        self.my_controller = None
        self.log = log
        self.log.info('starting zwave model')
        self.notifier = Notifier()
        self.app_replies_outstanding = 0
        self.suc = None
        self.client_command = ClientCommand()
        self.unsol_queue = Queue(50)
        self.last_update_ping = 0
        self.poll_alarms_soon = False
        self.wake_from_sleep = False
        self.mult_units = -1
        
    def set_serial(self, serial_api):
        self.network = ZWNetwork(serial_api)
        self.network.serial_api.set_model(self)
        cmd = SerialAPISoftReset(serial_api)
        cmd.go()

    def set_startup(self):
        self.startup = True

    
    def get_status(self):
        serial = 'unknown'
        if self.network and self.network.serial_api:
            serial = self.network.serial_api.get_state()
        return {'serial': serial, 'model': self.state}
    
    def is_usb_ok(self):
        if self.network and self.network.serial_api:
            return (self.network.serial_api.get_state() == 'started')
        else:
            return False
    
    def set_multiply(self, value):
        self.mult_units=value
        
    def notify(self, notice):
        """convenient method to send notifications"""
        self.notifier.send_notify('TrickleStar', notice)
        
    def build(self, polling=False):    
        """ construct the network model """
        logging.info("building network model")
        self.state = 'building'
        if self.network:
            self.network.start()
        else:
            raise ZWModelError, "no network interface available"
        self.poll_lock.acquire()
        try:
            # get an updated node list
            cmd = ZWSpecialVersion(self.network.serial_api)
            new_node_list = cmd.get_node_list()
            logging.info("node list is: " + ",".join([str(n) for n in new_node_list]))
            # find the controller we are connected though
            idcmd = MemoryGetID(self.network.serial_api)
            (home, node) = idcmd.get_ids()
            logging.info('usb stick at home_id %s node_id %s' % (hex(home),hex(node)))
            if  not check_id(home):
                logging.info('device not supported')
            if self.devices.has_key(node) and self.devices[node].is_my_controller():
                #alredy have this controller
                #check modes
                xcmd = ZWGetControllerCapabilities(self.network.serial_api)
                try:
                    self.capabilities = xcmd.get_capabilities()
                except:
                    self.log.error('USB stick get capabilities failed')
                    raise
                logging.info('usb controller capabilities are: ' + str(self.capabilities))
                usb_dev = self.devices[node]
            else:
                new_dev = ZWDevice(self, node)
                self.devices[node] = new_dev
                new_dev.make_usb_controller()
                usb_dev = new_dev
            usb_dev.home_id=home
            # find the suc stuff
            getid = ZWGetSUCNodeID(self.network.serial_api)
            try:
                suc_id = getid.get_id()
            except:
                logging.error('error getting SUC node id')
                self.log.error('error getting SUC node id')
                raise
            # populate device objects
            for node_id in new_node_list:
                if not self.devices.has_key(node_id):
                    # add device
                    new_dev = ZWDevice(self, node_id)
                    # and let it work itself out
                    self.devices[node_id] = new_dev
                    if new_dev.populate(self.network.serial_api):
                        logging.info('populated node %s' % hex(node_id))
                    else:
                        # missing unit noprotocols
                        logging.info('no protocols for node %s' % hex(node_id))
            if suc_id != 0 and self.devices.has_key(suc_id):
                logging.info('setting node %d as SUC' % suc_id)
                self.suc = self.devices[suc_id]
            else:
                logging.info('no suc available')
                self.suc = None
        except:
            self.poll_lock.release()
            self.last_poll_time = time.time()            
            raise
        else:
            self.poll_lock.release()
            self.last_poll_time = time.time()            
        self.state = 'running'
        #logging.debug('devices are ' + repr(self.devices))
        #for d in self.devices.keys():
        #    logging.debug('device node id: ' + str(d) + ' -- ' + str(self.devices[d].node_id))
                
    
    def can_send_app_cmd(self):
        return self.app_replies_outstanding == 0
    
    def poll(self, times=1, noapp=False):
        """ Poll the connection and handle incoming packets """
        # build if we have no units...
        if len(self.devices) == 0:
            self.build()
            return
        if self.mode == 'exclude_end':
            self.stop_remove()
        if self.mode == 'include_end':
            self.stop_add()
        try:
            unsol = self.unsol_queue.get_nowait()
            self.process_unsol(unsol)
        except Empty:
            pass
        except:
            logging.exception('unsol queue exception')
        if not noapp:
            # run scheduler and poll times
            self.schedule_and_net_poll()
            # send from queue if possible
            time.sleep(0.1)
            try:
                if True:
                    (nodeid, appcmd) = self.send_queue.get_nowait()
                    logging.debug('getting cmd from queue for node_id = %d' % int(nodeid))
                    if self.devices.has_key(nodeid):
                        dodev = self.devices[nodeid]
                        dodev.send_app_command(appcmd, False)
                        self.app_replies_outstanding=1
            except Empty:
                pass
            except:
                logging.exception('queue exception')
        for f in range(times):
            self.poll_lock.acquire()
            self.network.serial_api.lock.acquire()
            logging.debug('polling')
            got = self.network.serial_api.read_something()
            self.network.serial_api.lock.release()
            self.poll_lock.release()
            self.process_unsol(got)


    def queue_unsol(self, frame):
        """Queue unsolicited frame to be picked up by polling"""
        try:
            self.unsol_queue.put_nowait(frame)
        except:
            pass
        
        
        
    def process_unsol(self, got):
        if isinstance(got, ZWFrame):
            # there's a frame to process
            logging.debug('poll: got incoming frame')
            if got.command:
                if got.command[0] == 0x04:
                    logging.debug('poll: got app command')
                    try:
                        self.app_command_handler(got)
                    except Exception,e:
                        logging.warning(e)
                elif got.command[0] == 0x49:
                    logging.debug("poll: got unsol(?) node info frame")
                    try:
                        self.unsol_node_info_frame_handler(got)
                    except Exception,e:
                        logging.exception(e)
                elif got.command[0] == 0x4a:
                    self.device_add_response(got)
                elif got.command[0] == 0x4b:
                    self.device_remove_response(got)
                elif got.command[0] == 0x13:
                    self.tx_status_handler(got)
                elif got.command[0] == 0x50:
                    self.learn_response(got)
                else:
                    logging.warning('poll: received unknown frame cmd=' + hex(got.command[0]))
        


    def has_suc(self):
        """
        True if this network has a SUC
        """
        return self.suc != None
    
    def tx_status_handler(self,frame):
        """
        process tx status being returned 
        """
        if frame.is_res():
            # the response, will only have 
            logging.debug('zwsenddata returned %d' % frame.command[1])
        elif frame.is_req():
            # the request type -- so this is useful info
            # the frame should look like:
            # ZW->HOST: REQ | 0x13 | funcID | txStatus
            if len(frame.command) < 3:
                # hmm, incomplete
                logging.debug('got incomplete frame from zwsendata response')
            else:
                funcid = frame.command[1]
                tx_status = frame.command[2]
                logging.debug('got txstatus frame for %d with status %d' % (funcid, tx_status))
                self.app_replies_outstanding = 0
                this_fid = fid.lookup_fid(funcid)
                if this_fid and self.devices.has_key(this_fid.node_id):
                    mydev = self.devices[this_fid.node_id]
                else:
                    return
                if tx_status == TRANSMIT_COMPLETE_OK:
                    # successful transmit
                    mydev.tx_confirm()
                elif tx_status == TRANSMIT_COMPLETE_NO_ACK:
                    # remote unit did not respond
                    mydev.tx_no_ack()
                elif tx_status == TRANSMIT_COMPLETE_FAIL:
                    # soft failure
                    self.tx_net_fail()
                else:
                    pass
        else:
            pass
    
    def device_add_response(self, frame):
        """
        ZW->HOST: REQ | 0x4A | funcID | bStatus | bSource | bLen | basic | generic | specific | cmdclasses[]         
        [0x4a,0xaa,0x7,0x0,0x0]
        """
        logging.info('got add node frame (with subcommand %d)' % frame.command[2])
        if frame.command[1] == 0xaa:
            if (frame.command[2] == 0x03 or frame.command[2] == 0x04):
                new_ni = frame.command[3]
                logging.info('adding node ' + str(new_ni))
                # send done to this process first:
                # construct a protocol frame
                # make a new device
                clen = frame.command[4]
                if clen > 3:
                    basic = frame.command[5]
                    generic = frame.command[6]
                    specific = frame.command[7]
                    command_classes = [ frame.command[8+i] for i in range(clen-4) ]
                    p = Protocols(basic, generic, specific, command_classes)
                    new_dev = ZWDevice(self, new_ni)
                    self.devices[new_ni] = new_dev
                    # and let it work itself out
                    new_dev.attach_protcols(p)
                    logging.info('new node added')
                    self.client_command.add_command({'cmd': 'del_config', 'node': new_ni})
                    # if new device is a static controller and we are primary, then make it suc
                    if (basic == 0x02 and not self.has_suc()):
                        logging.info('asking new static controller node %d to be SUC' % new_ni)
                        # ask this controller to be suc
                        suc_cmd = ZWSetSUCNodeID(self.network.serial_api)
                        try:
                            suc_cmd.go(new_ni,capabilities=ZWSetSUCNodeID.ZW_SUC_FUNC_NODEID_SERVER)
                        except:
                            self.log.error('setsucnodeid failed')
                            raise
            elif frame.command[2] == 0x05:
                if self.mode == 'include':
                    self.mode = 'include_end'
                    self.mode_end = 0
            elif frame.command[2] == 0x06:
                if self.mode == 'include':
                    self.mode = 'include_end'
                    self.mode_end = 0
                logging.info('include stopped: status %d' % frame.command[2])
            elif frame.command[2] == 0x07:
                if self.mode == 'include':
                    self.mode='include_end'
                    self.mode_end = 0
                logging.info('include stopped: status %d' % frame.command[2])
                # add to error log
                self.log.error('include stopped (error 7).')
                
    def device_remove_response(self, frame):
        """
        ZW->HOST: REQ | 0x4B | funcID | bStatus | bSource | bLen | basic | generic | specific | cmdclasses[] 
        0x4b,0x9a,0x6,0x0,0x0
        [0x4b,0x9a,0x6,0x0,0x9,0x3,0x11,0x1,0x26,0x27,0x72,0x86,0x75,0x73]
        """
        logging.info('got device remove response')
        fid = frame.command[1]
        status = frame.command[2]
        old_ni = frame.command[3]
        blen = frame.command[4]
        if fid == 0x9a:
            if self.mode=='exclude' and (status == 0x03 or status == 0x04) and old_ni > 0:
                logging.info('removing node ' + str(old_ni))
                if self.devices.has_key(old_ni):
                    del self.devices[old_ni]
                self.client_command.add_command({'cmd': 'del_config', 'node': old_ni})
            elif status == 0x06:
                logging.info('exclude stopped: status %d' % status)
                if self.mode == "exclude":
                    self.mode="exclude_end"
                    self.mode_end = 0
            elif status == 0x07:
                logging.info('exclude stopped: status %d' % status)
                if self.mode=="exclude":
                    self.mode="exclude_end"
                    self.log.error('exclude stopped (error 7).')
                    self.mode_end = 0
        
    def learn_response(self, frame):
        """
        ZW->HOST: REQ | 0x50 | funcID | bStatus | bSource | bLen | pCmd[] 
        #define LEARN_MODE_STARTED                  1
        #define LEARN_MODE_DONE                     6
        #define LEARN_MODE_FAILED                   7
        #define LEARN_MODE_DELETED                  0x80
        """
        status = frame.command[2]
        logging.info('learn mode returned %d' % status)
        if status == 0x06:
            logging.info('joined other network, reloading')
            self.devices  = {}
            self.network_update_count = 0
            
    def app_command_handler(self, frame):
        """ despatch and app command of some kind
        ZW->HOST: REQ | 0x04 | rxStatus | sourceNode | cmdLength | pCmd[]
         """
        source_node = frame.command[2]
        cmd_len = frame.command[3]
        cmd = frame.command[4:4+cmd_len]
        logging.debug('app command %s from node %s' % (hex(cmd[0]), hex(source_node)))
        if self.devices.has_key(source_node):
            (self.devices[source_node]).app_command_handler(cmd)
        
        
    def unsol_node_info_frame_handler(self, frame):
        source_node = frame.command[2]
        stat = frame.command[1]
        cmd_len = frame.command[3]
        cmd = frame.command[4:4+cmd_len]
        if self.devices.has_key(source_node):
            (self.devices[source_node]).unsol_node_info_handler(cmd)
        else:
            if stat == UPDATE_STATE_NODE_INFO_RECEIVED:
                # not a node we've seen before... so add it?
                new_dev = ZWDevice(self, source_node)
                self.devices[source_node] = new_dev
                if new_dev.populate(self.network.serial_api):
                    logging.info('populated node %s' % hex(source_node))
                else:
                    # missing unit noprotocols
                    logging.info('no protocols for node %s' % hex(source_node))
            elif stat == UPDATE_STATE_SUC_ID:
                logging.info('suc id changed to %s' % hex(source_node))
                if not self.devices.has_key(source_node):
                    new_dev = ZWDevice(self, source_node)
                    self.devices[source_node] = new_dev
                    if new_dev.populate(self.network.serial_api):
                        logging.info('populated node %s' % hex(source_node))
                    else:
                        # missing unit noprotocols
                        logging.info('no protocols for node %s' % hex(source_node))
                self.suc = self.devices[source_node]
            else:
                logging.info('unsol nodeinfo for node = %s state=%s' % (hex(source_node), hex(stat)))
    def schedule_and_net_poll(self):
        # check to see if we need to schedule any events
        # this is on a per-device basis
        # and run once per minute
        
        # firstly, check for startup
        if self.startup:
            # process on startup rules
            self.system_startup()
            self.startup = False
        
        if self.wake_from_sleep:
            self.system_wake_from_sleep()
            
        #
        # manage timed modes
        # see if we need to turn off
        #
        if self.mode_end != 0 and self.mode_end < time.time():
            if self.mode == 'include':
                self.stop_add()
            elif self.mode == 'exclude':
                self.stop_remove()
        
        #
        # manage alarm polling
        # 
        if self.poll_alarms_soon:
            self.poll_alarms_soon = False
            for k in self.devices.keys():
                self.devices[k].poll_alarms()
        
        #
        # scheduler
        #
        
        t = time.localtime()
        if t.tm_min != self.last_minute:
            last = self.last_minute
            self.last_minute = t.tm_min
            logging.debug('running scheduler for minute %d' % t.tm_min)
            # run schedules
            for d in self.devices.keys():
                self.devices[d].run_schedule(t, last)
        # check for network polls
        self.roll_poll()
    
    def roll_poll(self):
        """
        We get this every minute, and here we check to see if it is our poll time and if so:
         * check node_id list
         * check the connection is still there
        
        """
        conf = get_config()
        # sort out the units list first
        
        now = time.time()
        
        # check to see if we need to do network update
        # only if we are secondary, inclusion, or primary with an SUC:
        if self.my_controller and (self.my_controller.is_inclusion_controller() or self.my_controller.is_secondary_controller() or (self.my_controller.is_primary_controller() and self.has_suc())):
            if (self.network_update_count < 4):
                delay = self.short_network_update_delay
            else:
                delay = self.network_update_delay
            if now > (self.last_network_update_time + delay):
                self.last_network_update_time = now
                # do network update
                self.request_network_update()
                self.network_update_count += 1
        
        #logging.debug('last poll at %d now %d polltime %d' % (self.last_poll_time, now, int(conf['polltime'])))
        #logging.debug('config is ' + repr(conf))
        if not conf.has_key('polltime'):
            conf['polltime'] = 1;
        if ((now - self.last_poll_time) / 60.0) >= int(conf['polltime']):
            # do unit poll etc
            logging.info('doing network poll')
            self.last_poll_time = time.time()
            self.build(polling=True)
        # but only if poll time
        # then check the devices -- have to roll their poll times
            for d in self.devices.keys():
                self.devices[d].roll_poll()
    
    
    def system_sleep(self):
        """
        The system is about to sleep. Go thorough devices and 
        give them the chance to do something before sleep
        
        """
        for k in self.devices.keys():
            self.devices[k].system_sleep()
            self.poll(noapp=True)

    def system_wake_from_sleep(self):
        """
        The system is about to sleep. Go thorough devices and 
        give them the chance to do something before sleep

        """
        for k in self.devices.keys():
            self.devices[k].system_wake_from_sleep()
            self.wake_from_sleep = False
        self.poll(noapp=True)

    def set_wake_from_sleep(self):
        self.wake_from_sleep = True
        
    def system_startup(self):
        """
        The system is about to sleep. Go thorough devices and 
        give them the chance to do something before sleep
        
        """
        for k in self.devices.keys():
            self.devices[k].system_startup()
            self.poll(noapp=True)
        
    
    def system_shutdown(self):
        """
        The system is about to sleep. Go thorough devices and 
        give them the chance to do something before sleep
        
        """
        for k in self.devices.keys():
            self.devices[k].system_shutdown()
            self.poll(noapp=True)
        
    def is_weekday(self, day):
        global config
        return config and config['weekdays'][day] 
    
    def is_weekend(self, day):
        global config
        return config and config['weekends'][day]
        
    def __str__(self):
        """ Print out the whole model somehow """
        node_ids = self.devices.keys()
        p = " Nodes at: " + ",".join([str(n) for n in node_ids])
        for n in node_ids:
            p += '\n' + str(self.devices[n])
        return p
    
    def get_nodelist(self):
        return self.devices.keys()
    
    def start_remove(self):
        logging.info('starting exclude mode')
        try:
            cmd = ZWRemoveNodeFromNetwork(self.network.serial_api)
            cmd.start_remove()
        except:
            self.log.error('can not start exclude')
            raise
        self.mode = 'exclude'
        self.mode_end = time.time() + self.mode_delay
        return [True]

    def stop_remove(self):
        logging.info('stopping exclude mode')
        try:
            cmd = ZWRemoveNodeFromNetwork(self.network.serial_api)
            cmd.stop_remove()
        except:
            self.log.error('can not stop exclude')
            raise
        self.mode = 'run'
        self.mode_end = 0
        return [True]

    def start_add(self, args=None):
        replicate = args and args.has_key('replicate_controller')
        if replicate:
            logging.info('starting controller replication')
        else:
            logging.info('starting include mode')
        try:
            cmd = ZWAddNodeToNetwork(self.network.serial_api)
            cmd.start_add()
        except:
            self.log.error('can not start include')
            raise
        self.mode = 'include'
        self.mode_end = time.time() + self.mode_delay
        return [True]

    def stop_add(self):
        logging.info('stopping include mode')
        try:
            cmd = ZWAddNodeToNetwork(self.network.serial_api)
            cmd.stop_add()
        except:
            self.log.error('can not stop include')
            raise
        self.mode = 'run'
        self.mode_end = 0
        return [True]
    
    def reset_controller(self):
        """reset the current controller"""
        try:
            cmd = ZWSetDefault(self.network.serial_api)
            cmd.go()
        except:
            self.log.error('reset of controller failed')
            raise
        self.devices={}
        return [True]
    
    def add_to_network(self):
        """
        add this controller to an existing network
        """
        # set learn mode
        try:
            cmd = ZWSetLearnMode(self.network.serial_api)
            cmd.start_learn()
        except:
            self.log.error('add to network failed')
            raise
        return [True]
        
    def request_network_update(self):
        """
        As secondary or inclusion controller, ask for network update
        
        """
        self.log.info('requesting network update')
        try:
            cmd = ZWRequestNetworkUpdate(self.network.serial_api)
            cmd.go()
            self.poll(15)
        except:
            self.log.error('request network update failed')
            raise
        self.devices = {}
        return [True]
        
        
    def update_ping(self):
        """called when an update seen, so we know when to refresh alarms etc"""
        now = time.time()
        if (now - self.last_update_ping) > 60:
            logging.debug('polling alarms')
            self.poll_alarms_soon = True
        self.last_update_ping = now
        
    def tx_net_fail(self):
        pass
    
    def mutiply_devices(self, devs):
        """Manufacture some extra devices to look good"""
        wanted = self.mult_units
        index = 0
        orig_length = len(devs)
        while wanted > len(devs):
            if index==0: index += 1
            ni = devs[index]['nodeid']
            sim_dev = self.devices[ni].get_info()
            sim_dev['nodeid'] = len(devs) + 33
            sim_dev['level'] = random.randint(0,99)
            devs.append(sim_dev)
            index = (index+1)%orig_length

class ZWNetwork(object):
    def __init__(self, serial_api):
        """ set up network serial interface """
        self.serial_api = serial_api
    
    def start(self):
        """start the network operating .. mostly port opening"""
        if self.serial_api:
            self.serial_api.start()
        else:
            # try and start the network
            self.serial_api = ZWSerialAPI()
            self.serial_api.start()
    
    def stop(self):
        """shut down the network"""
        if self.serial_api:
            self.serial_api.stop()

    
    def check(self):
        """check network is still working and open"""
        #TODO -- fill in check details here -- random command or something?
        return True



class ZWDevice(object):
    """
    A ZWave device -- that is specialized by device and command classes
    """
    def __init__(self, model, node_id):
        """  """
        self.model = model
        self.node_id = node_id
        self.home_id = 0

        self.base = None
        self.generic = None
        self.specific = None
        self.comclass = {}
        self.class_names = {}
        self.level = 0
        self.capabilities = None #controller capabilities
        self.config = get_device_config(node_id)
        self.log = DeviceLog()
        self.is_usb_controller = False;
        self.manufacturer_specific = ManufacturerSpecific()
        self.last_poll = time.localtime()
        self.comm_failure = False
        self.log.info('starting log')
        self.version = None
        self.ccversions = {}
        self.alarms={}
    
    def log_error(self, err):
        self.log.error(err)
        # alert command
        self.model.client_command.add_command({'cmd': 'show_error', 'node': self.node_id})
        # and notify
        # TODO
        
    def populate(self, serial):
        """
        query myself on network and attach behaviours
        """
        # query basics from EEPROM
        
        icmd = ZWGetNodeProtocolInfo(self.model.network.serial_api, self.node_id)
        try:
            current = icmd.get_protocols()
        except:
            self.log_error('can not get node protocol info')
            raise
        self.protocols = current
        self._attach_behaviours()  # first time, sort out the mandatory ones

        # and 

        cmd = ZWRequestNodeInfo(serial, self.node_id)
        try:
            self.protocols = cmd.get_protocols()
        except:
            self.log_error('request node info failed')
        if not self.protocols:
            # device didn't return protocols
            if not self.comm_failure:
                self.log_error('No response from device')
                self.comm_failure = True
        else:
            self._attach_behaviours()
            self.last_poll = time.localtime()
            self.comm_failure = False
        self.log.info('device created, classes attached')
        # need to poll device to get levels etc
        self.unsol_node_info_handler(None)
        return True
        #except Exception, e:
        #    logging.warning("populate failed for node: "+ str(self.node_id) + ":" + e.message + str(e))
    
    def attach_protcols(self, protocols):
        self.protocols = protocols
        self._attach_behaviours()

    def app_command_handler(self, command):
        """
            Called when incoming application command is found.
            Decode message and apply to correct command class
        """
        cclass = command[0]
        logging.debug('got command class %s command %s' % (hex(cclass), hex(command[1])))
        if self.comclass.has_key(cclass):
            self.comclass[cclass].app_command_handler(command)
        else:
            # add in the class so it can be used
            logging.debug('adding command class %s', hex(cclass))
            self.comclass[cclass] = (ZWCommandClasses.get_class(cclass))(self)
            self.comclass[cclass].app_command_handler(command)
        
    def unsol_node_info_handler(self, command):
        """
            We've received a node info message.  It might be
            either:  1. it has just shown up, or 2. we asked for it
            and it took a while to arrive
            
        """
        # make sure we know about these command classes, and support them if need be
        # in case we haven't seen em before
        
        # check for binary switch or multilevel -- in which case exclude basic
        exclude_basic_poll = False
        for cc in self.comclass.values():    
            if cc.name == 'COMMAND_CLASS_SWITCH_BINARY' or cc.name == 'COMMAND_CLASS_SWITCH_MULTILEVEL':
                logging.debug("node %d has binary or multilevel, igonring basic poll" % self.node_id)
                exclude_basic_poll = True
                break
        
        # do device updates for levels etc.
        for cc in self.comclass.values():
            if not (exclude_basic_poll and cc.name == 'COMMAND_CLASS_BASIC'):
                cc.unsol_node_info_handler(command)
            
    def add_queue(self, command):
        """docstring for add_queue"""
        logging.debug('adding cmd to queue for node_id=%d' % int(self.node_id))
        self.model.send_queue.put((self.node_id, command), False)

    def run(self, class_short_name, command, *args, **kw):
        """
        Run a command class command, for the given class and name
        """
        if not self.class_names.has_key(class_short_name):
            self.log_error('command class %s not available' % class_short_name)
            return [False]
            #raise ZWModelError, "command class " + class_short_name + " not found"
        # despach to the class itself
        return self.class_names[class_short_name].run(command, *args, **kw)
        
    def _attach_behaviours(self):
        """
        Attach class behaviours to this object based on zwave device and command classes
        """
        logging.info('protocols for node %d: %s' % (self.node_id, self.protocols))

        self.base = (ZWBasicDeviceClasses.get_class(self.protocols.basic))(self)
        gen_spec = (ZWGenericSpecificDeviceClasses.get_class(self.protocols.generic, self.protocols.specific))
        self.generic = gen_spec[0](self)
        self.specific = gen_spec[1](self)
        
        # add in mandatory support        
        mand_nums  = self.base.man_classes + self.generic.man_classes + self.specific.man_classes
        for cmd_class in mand_nums:
            self.comclass[cmd_class] = (ZWCommandClasses.get_class(cmd_class))(self)
        
        logging.info('adding mandatory command classes %s' % ", ".join([hex(m) for m in mand_nums]))
        # and may get replaced later..
        
        for cmd_class in self.protocols.command_classes:
            self.comclass[cmd_class] = (ZWCommandClasses.get_class(cmd_class))(self)
        # Make hash of short_names
        for cid in self.comclass.keys():
            self.class_names[self.comclass[cid].short_name] = self.comclass[cid]
        #self.class_names[self.base.short_name] = self.base
        #self.class_names[self.generic.short_name] = self.generic
        #self.class_names[self.specific.short_name] = self.specific

    def make_usb_controller(self):
        """
        Set up the USB controller properly with correct base, generic, specific and command
        classes.
        
        """
        self.protocols = Protocols(0x01, 0x02, 0x01, [0x20])
        self._attach_behaviours()
        self.setup_controller()
        # work out what sort of controller this is -- the network
        # needs to know this as well
        cmd = ZWGetControllerCapabilities(self.model.network.serial_api)
        try:
            self.capabilities = cmd.get_capabilities()
        except:
            self.log_error('USB stick get capabilities failed')
            raise
        logging.info('usb controller capabilities are: ' + str(self.capabilities))
        
    def setup_controller(self):
        """
        Set up the USB device as the right soft of controller
        Using ApplicationNodeInformation
        """
        # 1. Get information from the device
        
        cmd = ZWGetNodeProtocolInfo(self.model.network.serial_api, self.node_id)
        try:
            current = cmd.get_protocols()
        except:
            self.log_error('getting controller protocol info failed')
            raise
        # 2. Check is the same
        logging.info('usb stick has generic %d specific %s' % (current.generic, current.specific))
        if not (current.generic == 0x02 and current.specific == 0x01):
            # 3. If needed, send it, and update defaults

            logging.info('USB Stick needs base classes updating.. Updating and restarting')
        
            cmd = ApplicationNodeInformation(self.model.network.serial_api)
            try:
                cmd.go(False, False, 0x02, 0x01, [0x20,0x21])
            except:
                self.log_error('setting application node info to stick failed')
                raise
            # And then start again.. may be needed
            self.devices=[]
            self.model.reset_controller()
 
        self.is_usb_controller = True
        logging.debug('setting controller to device %d' % self.node_id)
        self.model.my_controller=self
        
        self.manufacturer_specific.set_ms([0x0, 0x6B], [0x0, 0x28], [0x0, 0x0])
        
        # check SUC is disabled for this controller
        logging.info('disabling suc for stick')
        unsuc =  ZWEnableSUC(self.model.network.serial_api)
        try:
            unsuc.enable_suc(False)
        except:
            self.log.warning('unable to disable SUC for controller')
            raise
            #TODO remove re-raise here
        
    def is_my_controller(self):
        """
        Return True if this device is my controller
        
        """
        return self.is_usb_controller
        
    def is_primary_controller(self):
        return self.is_usb_controller and self.capabilities and self.capabilities.is_real_primary()
    
    def is_secondary_controller(self):
        return self.is_usb_controller and self.capabilities and self.capabilities.is_secondary()
    
    def is_inclusion_controller(self):
        return self.is_usb_controller and self.capabilities and self.capabilities.is_inclusion()
    
    
    def send_app_command(self, data, get_extra=False):
        if not self.is_usb_controller:
            cmd = ZWSendData(self.model.network.serial_api, self.node_id, data)
            try:
                cmd.go(self)
            except:
                self.log_error('error sending app comand to node %d' % self.node_id)
            #TODO bit of a hack here
            if get_extra:
                cmd.frame.responses = []
                self.model.network.serial_api.read_something(cmd.frame)

    def go(self, args):
        verb = args['verb']
        if verb == 'info':
            return self.get_info()
        elif verb == 'ccmd':
            # work with command class and command
            
            return self.run(args['cclass'], args['ccmd'], level=args['level']);
        elif verb == 'set_config':
            # save json config for device
            self.config = args['config']
            logging.debug('set_config for node %d %s' % (self.node_id, repr(self.config)))
            save_config(self.model)
            return [True]
        elif verb == 'get_log':
            return self.log.get_log()
        elif verb == 'clear_log':
            self.log.clear()
            return self.log.get_log()
        elif verb == 'indicate':
            self.indicate()
            return [True]
        else:
            return None
            
    def get_info(self):
        #logging.debug('getting info for node %s', hex(self.node_id))
        if self.model.my_controller:
            hid = hex(self.model.my_controller.home_id)
        else:
            hid = hex(0)
        try:
            i = { 
                'nodeid': self.node_id, 
                'base': self._get_typename(self.base), 
                'generic': self._get_typename(self.generic), 
                'specific': self._get_typename(self.specific),
                'level': self.level,
                'comm_failure': self.comm_failure,
                'alarms': self.alarms,
                'homeid': hid
                }
            if self.class_names:
                i['command_classes'] = self.class_names.keys()
            else:
                i['command_classes'] = []
            if self.capabilities:
                i['controller_capabilities'] = str(self.capabilities)
            else:
                i['controller_capabilities'] = None
            # put together device keys for display
            i['details'] = self.get_detail_keys()
            return i
        except:
            logging.exception('getinfo failed')

    def _get_typename(self, it):
        if it:
            return it.name
        else:
            return 'none'
            
    def _get_short_name(self, it):
        if it:
            return it.short_name
        else:
            return 'none'
        
    def get_detail_keys(self):
        keys = {}
        if self.capabilities:
            keys['Controller'] = str(self.capabilities)
        keys['Base Class'] = self._get_short_name(self.base)
        keys['Generic Class'] = self._get_short_name(self.generic)
        keys['Specific Class'] = self._get_short_name(self.specific)
        keys['Cmd. Classes'] = ",".join(self.class_names.keys())
        keys['Manufacturer'] = self.manufacturer_specific.get_manufacturer_for_display()
        keys['Product Type'] = self.manufacturer_specific.get_product_type_for_display()
        keys['Product Id'] = self.manufacturer_specific.get_product_id_for_display()
        return keys
        
    def __str__(self):
        p = str(self.node_id) + ': Classes:'
        p +=  "\n  Base :    " + self.base.short_name
        p +=  "\n  Generic:  " + self.generic.short_name
        p +=  "\n  Specific: " + self.specific.short_name
        p +=  "\n  Command Classes: " + ",".join(self.class_names.keys())
        return p
    
    def roll_poll(self):
        for cc in self.comclass.values():
            cc.unsol_node_info_handler(None)
    
    def run_schedule(self, t, last_min):
        if self.config:
            if self.config.has_key('schedule'):
                # there is a schedule, so something to work with
                if self.model.is_weekday(t.tm_wday):
                    logging.debug(' schedule> is a weekday')
                    self.run_schedule_part(self.config['schedule']['weekdays'], t, last_min)
                if self.model.is_weekend(t.tm_wday):
                    logging.debug(' schedule> is a weekday')
                    self.run_schedule_part(self.config['schedule']['weekends'], t, last_min)
    
    def run_schedule_part(self, config, t, last_min):
        if config['on']:
                try:
                    stime = time.strptime(config['ontime'],'%H:%M')
                    if stime.tm_hour == t.tm_hour and stime.tm_min <= t.tm_min and stime.tm_min > last_min:
                        self.model.notify('Turning on node %d' % self.node_id)
                        if self.class_names.has_key('basic'):
                            self.class_names['basic'].level_set(255)
                        elif self.class_names.has_key('switch_multilevel'):
                            self.class_names['switch_multilevel'].level_set(255)
                        else:
                            pass
                except Exception, e:
                    logging.exception(e)
                
                
        if config['off']:
            try:
                stime = time.strptime(config['offtime'],'%H:%M')
                if stime.tm_hour == t.tm_hour and stime.tm_min <= t.tm_min and stime.tm_min > last_min:
                    self.model.notify('Turning on node %d' % self.node_id)
                    if self.class_names.has_key('basic'):
                        self.class_names['basic'].level_set(0)
                    elif self.class_names.has_key('switch_multilevel'):
                        self.class_names['switch_multilevel'].level_set(0)
                    else:
                        pass
            except Exception, e:
                logging.exception(e)
    
    
    def indicate(self):
        """
        Implement indicate..  flash something somehow on/off 1 second delay or something
        or if implements the Indicate app, use that
        """
        if self.class_names.has_key('indicator'):
            self.class_names['indicator'].flash();
        elif self.class_names.has_key('basic'):
            self.class_names['basic'].flash();
        

    def system_sleep(self):
        """
        Process any on-sleep events for this device
        . look at config, execute what is necessary
        """
        if self.config:
            logging.debug('have config for node %d %s', self.node_id, repr(self.config))
            if self.config.has_key('rules') and self.config['rules'].has_key('1'):
                if self.config['rules']['1']:
                    # then look for turning off application
                    # either the basic or the special
                    if self.class_names.has_key('switch_multilevel'):
                        self.class_names['switch_multilevel'].level_set(0)
                    elif self.class_names.has_key('switch_binary'):
                        self.class_names['switch_binary'].level_set(0)                        
                    elif self.class_names.has_key('basic'):
                        self.class_names['basic'].level_set(0)
                    else:
                        pass

    def system_wake_from_sleep(self):
        """
        Process any on-sleep events for this device
        . look at config, execute what is necessary
        """
        if self.config:
            logging.debug('have config for node %d %s', self.node_id, repr(self.config))
            if self.config.has_key('rules') and self.config['rules'].has_key('6'):
                if self.config['rules']['6']:
                    # check for application for this device
                    if self.class_names.has_key('switch_multilevel'):
                        self.class_names['switch_multilevel'].level_set(255)
                    elif self.class_names.has_key('switch_binary'):
                        self.class_names['switch_binary'].level_set(255)                        
                    elif self.class_names.has_key('basic'):
                        self.class_names['basic'].level_set(255)
                    else:
                        pass

        
    def system_startup(self):
        """
        
        """
        if self.config:
            logging.debug('have config for node %d %s', self.node_id, repr(self.config))
            if self.config.has_key('rules') and self.config['rules'].has_key('0'):
                if self.config['rules']['0']:
                    # then look for turning on application
                    # either the basic or the special
                    if self.class_names.has_key('switch_multilevel'):
                        self.class_names['switch_multilevel'].level_set(255)
                    elif self.class_names.has_key('switch_binary'):
                        self.class_names['switch_binary'].level_set(255)                        
                    elif self.class_names.has_key('basic'):
                        self.class_names['basic'].level_set(255)
                    else:
                        pass
        
    def system_shutdown(self):
        """
        
        """
        if self.config:
            logging.debug('have config for node %d %s', self.node_id, repr(self.config))
            if self.config.has_key('rules') and self.config['rules'].has_key('2'):
                if self.config['rules']['2']:
                    # then look for turning off application
                    # either the basic or the special
                    if self.class_names.has_key('switch_multilevel'):
                        self.class_names['switch_multilevel'].level_set(0)
                    elif self.class_names.has_key('switch_binary'):
                        self.class_names['switch_binary'].level_set(0)                        
                    elif self.class_names.has_key('basic'):
                        self.class_names['basic'].level_set(0)
                    else:
                        pass



    def tx_confirm(self):
        self.comm_failure=False
    
    def tx_no_ack(self):
        if not self.comm_failure:
            self.log_error('Communications failure, no ack for command')
            self.comm_failure=True


    
    def is_tricklestar(self):
        """Return True if this is a tricklestar device"""
        if self.manufacturer_specific:
            return self.manufacturer_specific.is_tricklestar()
        else:
            return False
        
    def poll_alarms(self):
        """
        poll alarms, if supported
        """
        if self.class_names.has_key('alarm') and self.is_tricklestar():
            self.class_names['alarm'].alarm_check(None)
    
    def set_alarm(self, alarm_type, alarm_value):
        """ Set alarm, and notify of alarm condition"""
        if alarm_type < 0 or alarm_type > 1:
            logging.info('got invalid alarm type %d for node %d' % (alarm_type, self.node_id))
        old_alarm = self.alarms[alarm_type]
        self.alarms[alarm_type] = alarm_value
        if old_alarm == 0 and alarm_value > 0:
            # new alarm
            self.log_error('%s failed.' % ALARM_TYPE_NAMES[alarm_type]);
            self.model.notify('Alarm: %s failed for node %d' % (ALARM_TYPE_NAMES[alarm_type], self.node_id))
        
class ZWModelError(exceptions.Exception):
    """general model operation error"""
    def __init__(self, args=None):
        args = args
        
    

class zwmodelTests(unittest.TestCase):
    def setUp(self):
        pass


if __name__ == '__main__':
    unittest.main()