#!/usr/bin/env python
# encoding: utf-8
"""
zwserialapi.py

Created by Graeme Sutherland on 2008-03-03.
Copyright (c) 2008 TrickleStar Limited. All rights reserved.
"""



import sys
import os
import time
import unittest
import exceptions
import logging
import threading
from serial import *
from zwconstants import *

class ZWSerialAPI(object):
    
    STARTED = 1;
    STOPPED = 0;
    device = '/dev/cu.usbserial'
    serial = None

    def __init__(self):
        self.state = self.STOPPED
        self.serial = None
        self.buf = '';
        self.lock = threading.Lock()
        self.model = None
    
    def set_model(self, model):
        self.model = model

    def set_device(self, device):
        """set pyserial object"""
        self.device = device
    
    def start(self):
        #logging.info( 'starting serial api connection')
        if not self.serial:
            self.start_serial_port()
        if self.serial:
            self._clear_connection()
            self.state = self.STARTED
        
    def start_serial_port(self):
        #logging.info('opening serial port (zwserialapi)')
        try:
            self.serial = Serial(
                         port = self.device,
                         baudrate=115200,         #baudrate
                         bytesize=EIGHTBITS,    #number of databits
                         parity=PARITY_NONE,    #enable parity checking
                         stopbits=STOPBITS_ONE, #number of stopbits
                         timeout=0.5,          #set a timeout value, None to wait forever
                         xonxoff=0,             #enable software flow control
                         rtscts=0,              #enable RTS/CTS flow control
                         writeTimeout=None,     #set a timeout for writes
                         dsrdtr=None,           #None: use rtscts setting, dsrdtr override if true or false
                         )
            self.serial.setRTS(on=0)
            self.serial.setDTR(on=0)
            self.state= self.STARTED
             
        except Exception, e:
            # oops, no serial interface
            #logging.warning('Cant open device: ' + self.device + ' - ' + str(e))
            self.serial = None
            self.state = self.STOPPED
            return

    def get_state(self):
        if self.state == 1:
            return 'started'
        else:
            return 'stopped'

    def stop(self):
        """disconnect"""
        try:
            self.serial.setRTS(on=1)
            self.serial.setDTR(on=1)
        except:
            pass
        try:
            self.serial.close()
            self.state = self.STOPPED
            self.serial = None
        except:
            pass

    def check_connection(self):
        return self.state == self.STARTED

    def send_frame(self, frame, wait_for_response=False, retries=10):
        if self.state == self.STARTED:
            acked=False
            frame.matched=False
            frame.ack=False
            frame.can=False
            retry_count=0
            while (not acked) and (retry_count < retries):              
                logging.debug( 'sending frame: ' + str(frame))
                self.lock.acquire()
                self._write(frame.to_string())
                if len(frame.match) > 0:
                    match_attempts = 0;
                    #TODO add timeout here to prevent lockups
                    while frame.matched < len(frame.match) and match_attempts < 15:
                        logging.debug('no match yet .. reading')
                        self.read_something(frame)
                        if frame.can:
                            time.sleep(0.2)
                            self.read_something()
                            frame.can = False
                            retry_count -= 1
                        match_attempts += 1
                else:
                    self.read_something(frame)
                    if frame.can:
                        time.sleep(0.2)
                        self.read_something()
                        frame.can = False
                        retry_count -= 1   
                self.lock.release() 
                acked = frame.ack == True
                retry_count += 1

        else:
            #todo exception here
            return
        if wait_for_response:
            return self.read_something(frame)
        else:
            return ''
    
    def _get_ack(self):
        r = self._read(1)
        if  len(r) > 0:
            if ord(r) == ACK:
                logging.debug( ' got ACK')
                return True
            elif ord(r) == NAK:
                return False
            elif ord(r) == CAN:
                time.sleep(0.25)
                return False
            else:
                # push back
                self._unread(r)
                return False

    
    def read_something(self, frame = None):
        """
        Read bytes until we get to the logical end of something, decoding 
        messages as we go.  Possible end point are ACK, NAK, SOH <message> CSUM or 
        CAN
        """
        r =self._read(1)
        if len(r) > 0:
            if ord(r) == ACK:
                if frame: frame.ack = True
            elif ord(r) == NAK:
                if frame: frame.nak = True
            elif ord(r) == CAN:
                if frame: frame.can = True
            elif ord(r) == SOF:
                try:
                    # get and decode frame, returned as a frame
                    length = ord(self._read(1))
                    frame_type = ord(self._read(1))
                    command = self._read(length-2)
                    csum = ord(self._read(1))
                except:
                    # read failure
                    logging.warning('Read failure, incomplete packet, NAKing')
                    self._write(chr(NAK))
                    return None
                f = ZWFrame()
                f.set_type(frame_type)
                f.set_command([ord(c) for c in command])
                try:
                    f.verify_checksum(csum)
                except:
                    self._write(chr(NAK))
                    return None
                self._write(chr(ACK))
                if frame:
                    # check for a match
                    matched = False
                    for m in frame.match:
                        if len(m) == 2 and m[0] == frame_type and m[1] == ord(command[0]):
                            # a match
                            m.append(f)
                            matched=True
                            frame.add_response(f)
                            frame.matched += 1
                            return f
                    if not matched:
                        # despatch elsewhere
                        self.model.queue_unsol(f)
                else:
                    return f
            else:
                return r
        return None

    def _clear_connection(self):
        """
        Clear any existing packets. Attempt to reset the port to some reasonable sort of state before
        continuing.
        """
        time.sleep(2)
        while (self.serial.inWaiting() > 0):
            self.read_something()
        #self._write("\x06")
        #self.serial.flushOutput()

    def _read(self, read_len):
        logging.debug( " reading up to %d bytes with %d in buf" % (read_len, len(self.buf)))
        got=''
        if len(self.buf) > 0:
            ##TODO FINISH HERE
            toget = 0
            available = len(self.buf)
            if available > read_len:
                toget = read_len
            else:
                toget = available
            got = self.buf[0:toget]
            self.buf = self.buf[toget:]
            logging.debug(' read(buf) [' + ','.join([hex(ord(c)) for c in got]) + ']')
            logging.debug(' buf now is [' + ','.join([hex(ord(c)) for c in self.buf]) + ']')
            if len(got) == read_len:
                return got
            else:
                read_len = read_len - len(got)
        r =  self.serial.read(read_len)
        logging.debug(' read [' + ','.join([hex(ord(c)) for c in r]) + ']')
        return got + r
        
    
    def _unread(self, bytes):
        self.buf += bytes
        
    
    def _write(self, bytes):
        logging.debug( 'writing [' + ','.join([hex(ord(c)) for c in bytes]) + ']')
        self.serial.write(bytes)
        self.serial.flush()
        


class ZWSerialAPITests(unittest.TestCase):
    def setUp(self):
        pass



class ZWFrame(object):
    """A ZW transmission frame"""
    
    def __init__(self, type_in=None):
        self.frame_type = 0
        self.length = 0
        self.checksum = 0xff
        self.command = []
        self.ack = False
        self.nak = False
        self.can = False
        self.sent = None
        self.responses = []
        if type_in != None:
            self.frame_type = type_in

    def set_type(self, type_in):
        self.frame_type = type_in

    def is_req(self):
        return self.frame_type == REQ
    
    def is_res(self):
        return self.frame_type == RES

    def set_command(self, command):
        if isinstance(command,list):
            self.command = command
        else:
            self.command = [command]
        self.length = len(self.command) + 2

    def _checksum(self):
        cs = 0xff
        cs ^= self.length
        cs ^= self.frame_type
        for c in self.command:
            cs ^= c
        self.checksum=cs & 0xff
    
    def to_list(self):
        """docstring for to_bytes"""
        self._checksum()
        return [SOF, self.length, self.frame_type] + self.command + [self.checksum]

    def to_string(self):
        """docstring for to_bytes"""
        cstr = ''
        for c in self.to_list():
            cstr += chr(c)
        return cstr
        
    def add_response(self, response_frame):
        """docstring for add_response"""
        self.responses.append(response_frame)
    
    def verify_checksum(self, checksum):
        self._checksum()
        if self.checksum != checksum:
            raise ZWChecksumError, 'checksum verify failed for frame ' + str(self)

    def get_response_byte(self,byte_index):
        """
            get byte of response, works for single byte values only
        """
        if len(self.responses) > 0:
            return self.responses[0].get_arg_byte(byte_index)
        else:
            raise ZWMessageError, 'no responses for this message'

    def get_response_string(self,offset,size):
        """
            get byte of response, works for single byte values only
        """
        return self.responses[0].get_arg_string(offset,size)
        
    def get_arg_byte(self, byte_index):
        return self.command[byte_index]
    
    def get_arg_string(self,offset,size):
        return ''.join([chr(c) for c in self.command[offset:offset+size]])
        
    def __str__(self):
        s = '[' + ','.join([hex(x) for x in self.to_list()]) + ']'
        if self.can: s = s + ' CAN '
        if self.ack: s = s + ' ACK '
        if self.nak: s = s + ' NAK '
        #for resp in self.responses:
        #    s = s + "\n resp: " + str(resp)
        return s


class ZWChecksumError(exceptions.Exception):
    def __init__(self, args=None):
        args = args

class ZWMessageError(exceptions.Exception):
    def __init__(self, args=None):
        args = args
        
class ZWFrameTests(unittest.TestCase):
    def setUp(self):
        pass
    
    def testframe(self):
        f = ZWFrame(REQ)
        f.set_command(0x07)
        print(str(f))
        "[0x1,0x3,0x0,0x7,0xfb]" == str(f)



if __name__ == '__main__':
    unittest.main()